@@ -3,8 +3,8 @@ graphmodel.py
 
 Defines a graphical model container class for reasoning about graphical models
 
-Version 0.0.1 (2015-09-28)
-(c) 2015 Alexander Ihler under the FreeBSD license; see license.txt for details.
+Version 0.1.1 (2022-04-06)
+(c) 2015-2021 Alexander Ihler under the FreeBSD license; see license.txt for details.
 """
 
 import operator as operator
@@ -17,11 +17,14 @@ inf = float('inf')
 from pyGMs.factor import *
 
 
+def _is2D(D):
+  try: next(iter(next(iter(D))))
+  except: return False
+  return True
 
 
 # Make a simple sorted set of factors, ordered by clique size, then lexicographical by scope
 def factorSet(it=None): return SortedSet(iterable=it,key=lambda f:'{:04d}.'.format(f.nvar)+str(f.vars)[1:-1])
-#def factorSet(it=None): return SortedListWithKey(iterable=it,key=lambda f:'{:04d}.'.format(f.nvar)+str(f.vars)[1:-1])
 # So, fs = factorSet([list]) will make sure:
 #   fs[0] is a minimal factor (smallest # of variables) and fs[-1] is a maximal factor (largest # of variables)
 
@@ -53,7 +56,7 @@ class GraphModel(object):
 
 
   def __repr__(self):
-    # TODO: finish
+    # TODO: add any details?
     return "Graphical model: {} vars, {} factors".format(self.nvar,self.nfactors)
 
   def __init__(self, factorList=None, copy=True, isLog=False):   
@@ -168,17 +171,19 @@ class GraphModel(object):
     """Evaluate F(x) = \prod_r f_r(x_r) for some (full) configuration x.
          If optional subset != None, uses *only* the factors in the Markov blanket of subset.
     """
+    if not _is2D(x): x=[x]
     factors = self.factors if subset==None else self.factorsWithAny(subset)
-    if self.isLog: return np.exp( sum( [ f.valueMap(x) for f in factors ] ) )
-    else:          return np.product( [ f.valueMap(x) for f in factors ] )
+    if self.isLog: return np.exp( np.sum( [[ f.valueMap(xx) for f in factors ] for xx in x] ,1) )
+    else:          return np.product( [[ f.valueMap(xx) for f in factors ] for xx in x] ,1)
 
   def logValue(self,x,subset=None): 
     """Evaluate log F(x) = \sum_r log f_r(x_r) for some (full) configuration x.
          If optional subset != None, uses *only* the factors in the Markov blanket of subset.
     """
+    if not _is2D(x): x=[x]
     factors = self.factors if subset==None else self.factorsWithAny(subset)
-    if self.isLog: return sum( [ f.valueMap(x) for f in factors ] ) 
-    else:          return sum( [ np.log(f.valueMap(x)) for f in factors ] )
+    if self.isLog: return np.sum( [[ f.valueMap(xx) for f in factors ] for xx in x] ,1)
+    else:          return np.sum( [[ np.log(f.valueMap(xx)) for f in factors] for xx in x] ,1)
 
   def isBinary(self):   # Check whether the model is binary (all variables binary)
     """Check whether the graphical model is binary (all variables <= 2 states)"""
@@ -196,7 +201,7 @@ class GraphModel(object):
 
   def isBN(self, tol=1e-6):  # Check whether the model is a valid Bayes Net
     """Check whether the graphical model is a valid Bayes net (one CPT per variable) """
-    topo_order = bnOrder(self)                              # TODO: allow user-provided order?
+    topo_order = bnOrder(self)                              # TODO: allow user-provided order and check?
     if topo_order is None: return False
     # Now check to make sure each factor is a CPT for its last variable
     pri = np.zeros((len(topo_order),))-1
@@ -331,161 +336,156 @@ class GraphModel(object):
     
 
 
-  def nxMarkovGraph(self, all_vars=False):
-    """Get networkx Graph object of the Markov graph of the model
-
-    Example:
-    >>> G = nxMarkovGraph(model)
-    >>> nx.draw(G)
-    """
-    import networkx as nx
-    G = nx.Graph()
-    G.add_nodes_from( [v.label for v in self.X if (all_vars or v.states > 1)] ) 
-    for f in self.factors:
-      for v1 in f.vars:
-        for v2 in f.vars:
-          if (v1 != v2) and (all_vars or (v1.states > 1 and v2.states > 1)): 
-            G.add_edge(v1.label,v2.label)
-    return G 
-    """ Plotting examples:
-    fig,ax=plt.subplots(1,2)
-    pos = nx.spring_layout(G) # so we can use same positions multiple times...
-    # use nodelist=[nodes-to-draw] to only show nodes in model
-    nx.draw(G,with_labels=True,labels={0:'0',1:'1',2:'2',3:'3',4:'4',5:'5',6:'6'},
-              node_color=[.3,.3,.3,.7,.7,.7,.7],vmin=0.0,vmax=1.0,pos=pos,ax=ax[0])
-    nx.draw(G,with_labels=True,labels={0:'0',1:'1',2:'2',3:'3',4:'4',5:'5',6:'6'},
-              node_color=[.3,.3,.3,.7,.7,.7,.7],vmin=0.0,vmax=1.0,pos=pos,ax=ax[1])
-    """
-
-  def drawMarkovGraph(self,**kwargs):
-    """Draw a Markov random field using networkx function calls
-
-    Args:
-      ``**kwargs``: remaining keyword arguments passed to networkx.draw()
-
-    Example:
-    >>> model.drawMarkovGraph( labels={0:'0', ... } )    # keyword args passed to networkx.draw()
-    """
-    # TODO: fix defaults; specify shape, size etc. consistent with FG version
-    import networkx as nx
-    G = nx.Graph()
-    G.add_nodes_from( [v.label for v in self.X if v.states > 1] )  # only non-trivial vars
-    for f in self.factors:
-      for v1 in f.vars:
-        for v2 in f.vars:
-          if (v1 != v2): G.add_edge(v1.label,v2.label)
-    kwargs['var_labels'] = kwargs.get('var_labels',{n:n for n in G.nodes()})
-    kwargs['labels'] = kwargs.get('labels', kwargs.get('var_labels',{}) )
-    nx.draw(G,**kwargs)
-    return G
-
-
-
-  def drawFactorGraph(self,var_color='w',factor_color=(.2,.2,.8),**kwargs):
-    """Draw a factorgraph using networkx function calls
-
-    Args:
-      var_color (str, tuple): networkx color descriptor for drawing variable nodes
-      factor_color (str, tuple): networkx color for drawing factor nodes
-      var_labels (dict): variable id to label string for variable nodes
-      factor_labels (dict): factor id to label string for factor nodes
-      ``**kwargs``: remaining keyword arguments passed to networkx.draw()
-
-    Example:
-    >>> model.drawFactorGraph( var_labels={0:'0', ... } )    # keyword args passed to networkx.draw()
-    """
-    # TODO: specify var/factor shape,size, position, etc.; return G? silent mode?
-    import networkx as nx
-    G = nx.Graph()
-    vNodes = [v.label for v in self.X if v.states > 1]   # list only non-trivial variables
-    fNodes = [-i-1 for i in range(len(self.factors))]    # use negative IDs for factors
-    G.add_nodes_from( vNodes )
-    G.add_nodes_from( fNodes )
-    for i,f in enumerate(self.factors):
-      for v1 in f.vars:
-        G.add_edge(v1.label,-i-1)
-
-    if not 'pos' in kwargs: kwargs['pos'] = nx.spring_layout(G) # so we can use same positions multiple times...
-    kwargs['var_labels']  = kwargs.get('var_labels',{n:n for n in vNodes})
-    kwargs['labels'] = kwargs.get('var_labels',{})
-    nx.draw_networkx(G, nodelist=vNodes,node_color=var_color,**kwargs)
-    kwargs['labels'] = kwargs.get('factor_labels',{})    # TODO: need to transform?
-    nx.draw_networkx_nodes(G, nodelist=fNodes,node_color=factor_color,node_shape='s',**kwargs)
-    nx.draw_networkx_edges(G,**kwargs)
-    return G
-
-
-
-  def drawBayesNet(self,**kwargs):
-    """Draw a Bayesian Network (directed acyclic graph) using networkx function calls
-
-    Args:
-      ``**kwargs``: remaining keyword arguments passed to networkx.draw()
-
-    Example:
-    >>> model.drawBayesNet( labels={0:'0', ... } )    # keyword args passed to networkx.draw()
-    """
-    import networkx as nx
-    topo_order = bnOrder(self)                              # TODO: allow user-provided order?
-    if topo_order is None: raise ValueError('Topo order not found; graph is not a Bayes Net?')
-    pri = np.zeros((len(topo_order),))-1
-    pri[topo_order] = np.arange(len(topo_order))
-    G = nx.DiGraph()
-    G.add_nodes_from( [v.label for v in self.X if v.states > 1] )  # only non-trivial vars
-    for f in self.factors:
-      v2label = topo_order[ int(max(pri[v.label] for v in f.vars)) ]
-      for v1 in f.vars:
-        if (v1.label != v2label): G.add_edge(v1.label,v2label)
-
-    kwargs['var_labels'] = kwargs.get('var_labels',{n:n for n in [v.label for v in self.X]})
-    kwargs['labels'] = kwargs.get('labels', kwargs.get('var_labels',{}) )
-    kwargs['arrowstyle'] = kwargs.get('arrowstyle','->')
-    kwargs['arrowsize'] = kwargs.get('arrowsize',10)
-    nx.draw(G,**kwargs)
-    return G
-
-
-  def drawLimid(self, C,D,U, **kwargs):
-    """Draw a limited-memory influence diagram (limid) using networkx 
-
-    Args:
-      ``**kwargs``: remaining keyword arguments passed to networkx.draw()
-
-    Example:
-    >>> model.drawLimid(C,D,U, var_labels={0:'0', ... } )    # keyword args passed to networkx.draw()
-    """
-    import networkx as nx
-    decisions = [d[-1] for d in D]                       # list the decision variables
-    model = GraphModel( C + [Factor(d,1.) for d in D] )  # get all chance & decision vars, arcs
-    chance = [c for c in model.X if c not in decisions]
-    util = [-i-1 for i,u in enumerate(U)]
-    cpd_edges, util_edges, info_edges = [],[],[]
-    topo_order = bnOrder(model)                          
-    if topo_order is None: raise ValueError('Topo order not found; graph is not a Bayes Net?')
-    pri = np.zeros((len(topo_order),))-1
-    pri[topo_order] = np.arange(len(topo_order))
-    G = nx.DiGraph()
-    G.add_nodes_from( [v.label for v in self.X if v.states > 1] )  # only non-trivial vars
-    G.add_nodes_from( util )                                       # add utility nodes
-    for f in model.factors:
-      v2label = topo_order[ int(max(pri[v.label] for v in f.vars)) ]
-      for v1 in f.vars:
-        if (v1.label != v2label): 
-          G.add_edge(v1.label,v2label)
-          if v2label in decisions: info_edges.append((v1.label,v2label))
-          else: cpd_edges.append((v1.label,v2label))
-    for i,u in enumerate(U):
-      for v1 in u.vars:
-        G.add_edge(v1.label,-i-1)
-        util_edges.append( (v1.label,-i-1) )
-    if not 'pos' in kwargs: kwargs['pos'] = nx.spring_layout(G) # so we can use same positions multiple times...
-    nx.draw_networkx_nodes(G, nodelist=decisions, node_color=(.7,.7,.9), node_shape='s', **kwargs)
-    nx.draw_networkx_nodes(G, nodelist=chance, node_color=(1.,1.,1.), node_shape='o', **kwargs)
-    nx.draw_networkx_nodes(G, nodelist=util, node_color=(.7,.9,.7), node_shape='d', **kwargs)
-    nx.draw_networkx_edges(G, edgelist=cpd_edges+util_edges, **kwargs)
-    tmp = nx.draw_networkx_edges(G, edgelist=info_edges, **kwargs)
-    for line in tmp: line.set_linestyle('dashed')
-    nx.draw_networkx_labels(G, **kwargs)
+# def nxMarkovGraph(self, all_vars=False):
+#   """Get networkx Graph object of the Markov graph of the model
+#
+#   Example:
+#   >>> G = nxMarkovGraph(model)
+#   >>> nx.draw(G)
+#   """
+#   import networkx as nx
+#    G = nx.Graph()
+#    G.add_nodes_from( [v.label for v in self.X if (all_vars or v.states > 1)] ) 
+#    for f in self.factors:
+#      for v1 in f.vars:
+#        for v2 in f.vars:
+#          if (v1 != v2) and (all_vars or (v1.states > 1 and v2.states > 1)): 
+#            G.add_edge(v1.label,v2.label)
+#    return G 
+#    """ Plotting examples:
+#    fig,ax=plt.subplots(1,2)
+#    pos = nx.spring_layout(G) # so we can use same positions multiple times...
+#    # use nodelist=[nodes-to-draw] to only show nodes in model
+#    nx.draw(G,with_labels=True,labels={0:'0',1:'1',2:'2',3:'3',4:'4',5:'5',6:'6'},
+#              node_color=[.3,.3,.3,.7,.7,.7,.7],vmin=0.0,vmax=1.0,pos=pos,ax=ax[0])
+#    nx.draw(G,with_labels=True,labels={0:'0',1:'1',2:'2',3:'3',4:'4',5:'5',6:'6'},
+#              node_color=[.3,.3,.3,.7,.7,.7,.7],vmin=0.0,vmax=1.0,pos=pos,ax=ax[1])
+#    """
+#
+#  def drawMarkovGraph(self,**kwargs):
+#    """Draw a Markov random field using networkx function calls
+#
+#    Args:
+#      ``**kwargs``: remaining keyword arguments passed to networkx.draw()
+#
+#    Example:
+#    >>> model.drawMarkovGraph( labels={0:'0', ... } )    # keyword args passed to networkx.draw()
+#    """
+#    # TODO: fix defaults; specify shape, size etc. consistent with FG version
+#    import networkx as nx
+#    G = self.nxMarkovGraph()
+#    kwargs['var_labels'] = kwargs.get('var_labels',{n:n for n in G.nodes()})
+#    kwargs['labels'] = kwargs.get('labels', kwargs.get('var_labels',{}) )
+#    nx.draw(G,**kwargs)
+#    return G
+#
+#
+#
+#  def drawFactorGraph(self,var_color='w',factor_color=[(.2,.2,.8)],**kwargs):
+#    """Draw a factorgraph using networkx function calls
+#
+#    Args:
+#      var_color (str, tuple): networkx color descriptor for drawing variable nodes
+#      factor_color (str, tuple list): networkx color for drawing factor nodes
+#      var_labels (dict): variable id to label string for variable nodes
+#      factor_labels (dict): factor id to label string for factor nodes
+#      ``**kwargs``: remaining keyword arguments passed to networkx.draw()
+#
+#    Example:
+#    >>> model.drawFactorGraph( var_labels={0:'0', ... } )    # keyword args passed to networkx.draw()
+#    """
+#    # TODO: specify var/factor shape,size, position, etc.; return G? silent mode?
+#    import networkx as nx
+#    G = nx.Graph()
+#    vNodes = [v.label for v in self.X if v.states > 1]   # list only non-trivial variables
+#    fNodes = [-i-1 for i in range(len(self.factors))]    # use negative IDs for factors
+#    G.add_nodes_from( vNodes )
+#    G.add_nodes_from( fNodes )
+#    for i,f in enumerate(self.factors):
+#      for v1 in f.vars:
+#        G.add_edge(v1.label,-i-1)
+#
+#    if not 'pos' in kwargs: kwargs['pos'] = nx.spring_layout(G) # so we can use same positions multiple times...
+#    kwargs['var_labels']  = kwargs.get('var_labels',{n:n for n in vNodes})
+#    kwargs['labels'] = kwargs.get('var_labels',{})
+#    nx.draw_networkx(G, nodelist=vNodes,node_color=var_color,**kwargs)
+#    kwargs['labels'] = kwargs.get('factor_labels',{})    # TODO: need to transform?
+#    nx.draw_networkx_nodes(G, nodelist=fNodes,node_color=factor_color,node_shape='s',**kwargs)
+#    nx.draw_networkx_edges(G,**kwargs)
+#    return G
+#
+#
+#
+#  def drawBayesNet(self,**kwargs):
+#    """Draw a Bayesian Network (directed acyclic graph) using networkx function calls
+#
+#    Args:
+#      ``**kwargs``: remaining keyword arguments passed to networkx.draw()
+#
+#    Example:
+#    >>> model.drawBayesNet( labels={0:'0', ... } )    # keyword args passed to networkx.draw()
+#    """
+#    import networkx as nx
+#    topo_order = bnOrder(self)                              # TODO: allow user-provided order?
+#    if topo_order is None: raise ValueError('Topo order not found; graph is not a Bayes Net?')
+#    pri = np.zeros((len(topo_order),))-1
+#    pri[topo_order] = np.arange(len(topo_order))
+#    G = nx.DiGraph()
+#    G.add_nodes_from( [v.label for v in self.X if v.states > 1] )  # only non-trivial vars
+#    for f in self.factors:
+#      v2label = topo_order[ int(max(pri[v.label] for v in f.vars)) ]
+#      for v1 in f.vars:
+#        if (v1.label != v2label): G.add_edge(v1.label,v2label)
+#
+#    kwargs['var_labels'] = kwargs.get('var_labels',{n:n for n in [v.label for v in self.X]})
+#    kwargs['labels'] = kwargs.get('labels', kwargs.get('var_labels',{}) )
+#    kwargs['arrowstyle'] = kwargs.get('arrowstyle','->')
+#    kwargs['arrowsize'] = kwargs.get('arrowsize',10)
+#    nx.draw(G,**kwargs)
+#    return G
+#
+#
+#  def drawLimid(self, C,D,U, **kwargs):
+#    """Draw a limited-memory influence diagram (limid) using networkx 
+#
+#    Args:
+#      ``**kwargs``: remaining keyword arguments passed to networkx.draw()
+#
+#    Example:
+#    >>> model.drawLimid(C,D,U, var_labels={0:'0', ... } )    # keyword args passed to networkx.draw()
+#    """
+#    import networkx as nx
+#    decisions = [d[-1] for d in D]                       # list the decision variables
+#    model = GraphModel( C + [Factor(d,1.) for d in D] )  # get all chance & decision vars, arcs
+#    chance = [c for c in model.X if c not in decisions]
+#    util = [-i-1 for i,u in enumerate(U)]
+#    cpd_edges, util_edges, info_edges = [],[],[]
+#    topo_order = bnOrder(model)                          
+#    if topo_order is None: raise ValueError('Topo order not found; graph is not a Bayes Net?')
+#    pri = np.zeros((len(topo_order),))-1
+#    pri[topo_order] = np.arange(len(topo_order))
+#    G = nx.DiGraph()
+#    G.add_nodes_from( [v.label for v in self.X if v.states > 1] )  # only non-trivial vars
+#    G.add_nodes_from( util )                                       # add utility nodes
+#    for f in model.factors:
+#      v2label = topo_order[ int(max(pri[v.label] for v in f.vars)) ]
+#      for v1 in f.vars:
+#        if (v1.label != v2label): 
+#          G.add_edge(v1.label,v2label)
+#          if v2label in decisions: info_edges.append((v1.label,v2label))
+#          else: cpd_edges.append((v1.label,v2label))
+#    for i,u in enumerate(U):
+#      for v1 in u.vars:
+#        G.add_edge(v1.label,-i-1)
+#        util_edges.append( (v1.label,-i-1) )
+#    if not 'pos' in kwargs: kwargs['pos'] = nx.spring_layout(G) # so we can use same positions multiple times...
+#    nx.draw_networkx_nodes(G, nodelist=decisions, node_color=[(.7,.7,.9)], node_shape='s', **kwargs)
+#    nx.draw_networkx_nodes(G, nodelist=chance, node_color=[(1.,1.,1.)], node_shape='o', **kwargs)
+#    nx.draw_networkx_nodes(G, nodelist=util, node_color=[(.7,.9,.7)], node_shape='d', **kwargs)
+#    nx.draw_networkx_edges(G, edgelist=cpd_edges+util_edges, **kwargs)
+#    tmp = nx.draw_networkx_edges(G, edgelist=info_edges, **kwargs)
+#    for line in tmp: line.set_linestyle('dashed')
+#    nx.draw_networkx_labels(G, **kwargs)
 
 
 
@@ -853,6 +853,7 @@ def factorOrder(factors, varOrder):
 #def forwardSample(model, varOrder, factorOrder=None):
 def sampleSequential(model, varOrder, factorOrder=None):
   """Draw a configuration X sequentially from the model factors.
+  For a Bayes net using a valid topological order, this is equivalent to forward sampling.
  
   Returns logQ,xs  (tuple): the sampled config xs and its log-probability under the sampling distribution
   """
